# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/lacss.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/lacss.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from biopb.lacss import detection_request_pb2 as biopb_dot_lacss_dot_detection__request__pb2
from biopb.lacss import detection_response_pb2 as biopb_dot_lacss_dot_detection__response__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x62iopb/lacss/lacss.proto\x12\x0b\x62iopb.lacss\x1a#biopb/lacss/detection_request.proto\x1a$biopb/lacss/detection_response.proto2\xaf\x01\n\x05Lacss\x12M\n\x0cRunDetection\x12\x1d.biopb.lacss.DetectionRequest\x1a\x1e.biopb.lacss.DetectionResponse\x12W\n\x12RunDetectionStream\x12\x1d.biopb.lacss.DetectionRequest\x1a\x1e.biopb.lacss.DetectionResponse(\x01\x30\x01\x42\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.lacss_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_LACSS']._serialized_start=116
  _globals['_LACSS']._serialized_end=291
# @@protoc_insertion_point(module_scope)
