# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/bindata.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/bindata.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x62iopb/lacss/bindata.proto\x12\x0b\x62iopb.lacss\"o\n\x07\x42inData\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x33\n\nendianness\x18\x02 \x01(\x0e\x32\x1f.biopb.lacss.BinData.Endianness\"!\n\nEndianness\x12\x07\n\x03\x42IG\x10\x00\x12\n\n\x06LITTLE\x10\x01\x42\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.bindata_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_BINDATA']._serialized_start=42
  _globals['_BINDATA']._serialized_end=153
  _globals['_BINDATA_ENDIANNESS']._serialized_start=120
  _globals['_BINDATA_ENDIANNESS']._serialized_end=153
# @@protoc_insertion_point(module_scope)
