# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/detection_response.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/detection_response.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from biopb.lacss import roi_pb2 as biopb_dot_lacss_dot_roi__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$biopb/lacss/detection_response.proto\x12\x0b\x62iopb.lacss\x1a\x15\x62iopb/lacss/roi.proto\"9\n\tScoredROI\x12\x1d\n\x03roi\x18\x01 \x01(\x0b\x32\x10.biopb.lacss.ROI\x12\r\n\x05score\x18\x02 \x01(\x02\"?\n\x11\x44\x65tectionResponse\x12*\n\ndetections\x18\x01 \x03(\x0b\x32\x16.biopb.lacss.ScoredROIB\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.detection_response_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_SCOREDROI']._serialized_start=76
  _globals['_SCOREDROI']._serialized_end=133
  _globals['_DETECTIONRESPONSE']._serialized_start=135
  _globals['_DETECTIONRESPONSE']._serialized_end=198
# @@protoc_insertion_point(module_scope)
