# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/image_data.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/image_data.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from biopb.lacss import roi_pb2 as biopb_dot_lacss_dot_roi__pb2
from biopb.lacss import bindata_pb2 as biopb_dot_lacss_dot_bindata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1c\x62iopb/lacss/image_data.proto\x12\x0b\x62iopb.lacss\x1a\x15\x62iopb/lacss/roi.proto\x1a\x19\x62iopb/lacss/bindata.proto\"\xa6\x03\n\x06Pixels\x12%\n\x07\x62indata\x18\x03 \x01(\x0b\x32\x14.biopb.lacss.BinData\x12\x17\n\x0f\x64imension_order\x18\x04 \x01(\t\x12\r\n\x05\x64type\x18\x05 \x01(\t\x12\x0e\n\x06size_x\x18\t \x01(\r\x12\x0e\n\x06size_y\x18\n \x01(\r\x12\x0e\n\x06size_z\x18\x0b \x01(\r\x12\x0e\n\x06size_c\x18\x0c \x01(\r\x12\x0e\n\x06size_t\x18\r \x01(\r\x12\x1c\n\x0fphysical_size_x\x18\x0e \x01(\x02H\x00\x88\x01\x01\x12\x1c\n\x0fphysical_size_y\x18\x0f \x01(\x02H\x01\x88\x01\x01\x12\x1c\n\x0fphysical_size_z\x18\x10 \x01(\x02H\x02\x88\x01\x01\x12\x1a\n\x12physical_size_unit\x18\x11 \x01(\t\x12\x1b\n\x0etime_increment\x18\x12 \x01(\x02H\x03\x88\x01\x01\x12\x1b\n\x13time_increment_unit\x18\x15 \x01(\tB\x12\n\x10_physical_size_xB\x12\n\x10_physical_size_yB\x12\n\x10_physical_size_zB\x11\n\x0f_time_increment\"1\n\x0fImageAnnotation\x12\x1e\n\x04rois\x18\x01 \x03(\x0b\x32\x10.biopb.lacss.ROI\"h\n\tImageData\x12#\n\x06pixels\x18\x01 \x01(\x0b\x32\x13.biopb.lacss.Pixels\x12\x36\n\x10image_annotation\x18\x02 \x01(\x0b\x32\x1c.biopb.lacss.ImageAnnotationB\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.image_data_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_PIXELS']._serialized_start=96
  _globals['_PIXELS']._serialized_end=518
  _globals['_IMAGEANNOTATION']._serialized_start=520
  _globals['_IMAGEANNOTATION']._serialized_end=569
  _globals['_IMAGEDATA']._serialized_start=571
  _globals['_IMAGEDATA']._serialized_end=675
# @@protoc_insertion_point(module_scope)
